/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#define LOG_FRAME_DROPS false

#include "ExpCamera.h"

#include "MixCast.h"
#include "MixCastSdkActor.h"
#include "Data/SharedTexIds.h"
#include "Transmission/Thrift/ServiceCommunicator.h"
#include "Transmission/Thrift/ThriftUtil.hpp"
#include "Utility/MathUtil.hpp"
#include "ExpFrameSender.hpp"

#include "Engine/Engine.h"
#include "Camera/CameraComponent.h"
#include "Components/DrawFrustumComponent.h"
#include "Components/SceneCaptureComponent2D.h"
#include "Kismet/GameplayStatics.h"
#include "Runtime/Launch/Resources/Version.h"
#include "TimerManager.h"

#include <string>

AExpCamera::FFrameStartedEvent AExpCamera::FrameStarted;
AExpCamera::FFrameEndedEvent AExpCamera::FrameEnded;

AExpCamera::AExpCamera(const FObjectInitializer &ObjectInitializer)
{
	UPluginProjectSettings* ProjectSettings = SdkData::GetProjectSettings();

	SpawnCollisionHandlingMethod = ESpawnActorCollisionHandlingMethod::AlwaysSpawn;

	CameraTransform = CreateDefaultSubobject<USceneComponent>(FName("Transform"));
	SetRootComponent(CameraTransform);

	FullCapture = CreateMainCaptureComponent(false);
	if (ProjectSettings->CanRenderTransparentBg && ProjectSettings->TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
		FullAlphaCapture = CreateAlphaCaptureComponent(false);

	ForegroundCapture = CreateMainCaptureComponent(true);
	if (ProjectSettings->TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
		ForegroundAlphaCapture = CreateAlphaCaptureComponent(true);
}

void AExpCamera::BeginPlay()
{
	Super::BeginPlay();

	sdkModule = (FMixCast*)&IMixCast::Get();

	if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::PerPixel) {
		ForegroundCutoffMesh = NewObject<UForegroundCutoffMeshComponent>(this);
		ForegroundCutoffMesh->Rename(TEXT("Foreground Cutoff"));
		ForegroundCutoffMesh->AttachToCamera(this);
		ForegroundCutoffMesh->RegisterComponent();
	}

	VirtualCamera* camInfo = SdkData::GetCameraById(Identifier);

	timer = new FFrameTimer();

	framePipe = new FExpFrameSender(this);

	CreateLayerCanvas(camInfo);
}
USceneCaptureComponent2D* AExpCamera::CreateMainCaptureComponent(bool isForeground)
{
	FName compName(isForeground ? "Foreground" : "Full");
	USceneCaptureComponent2D* comp = CreateDefaultSubobject<USceneCaptureComponent2D>(compName);
	comp->SetupAttachment(CameraTransform, NAME_None);
	comp->bCaptureEveryFrame = false;
	comp->bCaptureOnMovement = false;
	comp->bAlwaysPersistRenderingState = true;
	comp->CaptureSource = ESceneCaptureSource::SCS_FinalColorLDR;
	comp->PrimitiveRenderMode = ESceneCapturePrimitiveRenderMode::PRM_RenderScenePrimitives;
	if (isForeground)
	{
		comp->ShowFlags.SetFog(false);
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 23
		comp->ShowFlags.SetAtmosphericFog(false);
#endif
		if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::Planar)
			comp->bEnableClipPlane = true;
	}
	return comp;
}
USceneCaptureComponent2D* AExpCamera::CreateAlphaCaptureComponent(bool isForeground)
{
	FName compName(isForeground ? "Foreground (Alpha)" : "Full (Alpha)");
	USceneCaptureComponent2D* comp = CreateDefaultSubobject<USceneCaptureComponent2D>(compName);
	comp->SetupAttachment(CameraTransform, NAME_None);
	comp->bCaptureEveryFrame = false;
	comp->bCaptureOnMovement = false;
	comp->bAlwaysPersistRenderingState = true;
	comp->CaptureSource = ESceneCaptureSource::SCS_SceneColorHDR;
	comp->PrimitiveRenderMode = ESceneCapturePrimitiveRenderMode::PRM_RenderScenePrimitives;
	comp->ShowFlags.SetFog(false);
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 23
	comp->ShowFlags.SetAtmosphericFog(false);
#endif
	comp->ShowFlags.SetLighting(false);
	comp->PostProcessBlendWeight = 0;
	if (isForeground)
	{
		if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::Planar)
			comp->bEnableClipPlane = true;
	}
	return comp;
}
void AExpCamera::CreateLayerCanvas(VirtualCamera* camInfo)
{
	LayerCanvas = UMixCastFrameCanvas::Create(this, camInfo->renderResolutionWidth, camInfo->renderResolutionHeight);

	FullCapture->TextureTarget = LayerCanvas->GetFullTex();
	if (FullAlphaCapture)
	{
		FullAlphaCapture->TextureTarget = LayerCanvas->GetFullAlphaTex();
	}

	ForegroundCapture->TextureTarget = LayerCanvas->GetForegroundTex();
	if (ForegroundAlphaCapture)
	{
		ForegroundAlphaCapture->TextureTarget = LayerCanvas->GetForegroundAlphaTex();
	}
}
void AExpCamera::DestroyLayerCanvas()
{
	if (LayerCanvas) {
		LayerCanvas->ConditionalBeginDestroy();
		LayerCanvas = nullptr;
	}
}

void AExpCamera::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	if (framePipe) {
		delete framePipe;
		framePipe = nullptr;
	}

	if (FullCapture) {
		if (FullCapture->IsValidLowLevel()) {
			FullCapture->TextureTarget = nullptr;
			FullCapture->DestroyComponent();
		}
		FullCapture = nullptr;
	}
	if (FullAlphaCapture) {
		if (FullAlphaCapture->IsValidLowLevel()) {
			FullAlphaCapture->TextureTarget = nullptr;
			FullAlphaCapture->DestroyComponent();
		}
		FullAlphaCapture = nullptr;
	}

	if (ForegroundCapture) {
		if (ForegroundCapture->IsValidLowLevel()) {
			ForegroundCapture->TextureTarget = nullptr;
			ForegroundCapture->DestroyComponent();
		}
		ForegroundCapture = nullptr;
	}

	if (ForegroundAlphaCapture) {
		if (ForegroundAlphaCapture->IsValidLowLevel()) {
			ForegroundAlphaCapture->TextureTarget = nullptr;
			ForegroundAlphaCapture->DestroyComponent();
		}
		ForegroundAlphaCapture = nullptr;
	}

	DestroyLayerCanvas();

	Super::EndPlay(EndPlayReason);
}

void AExpCamera::CaptureIfNeeded()
{
	VirtualCamera* camData = SdkData::GetCameraById(Identifier);
	if (camData == nullptr)
		return;

	if (!timer->IsStarted() || timer->GetFrameRate() != camData->renderFramerate)
		timer->Start(camData->renderFramerate);

	unsigned long oldFrameIndex = timer->LastFrameIndex();
	unsigned long long oldSyncTime = timer->LastIdealTime();
	timer->Update();
	unsigned long elapsedFrames = timer->LastFrameIndex() - oldFrameIndex;
	if (elapsedFrames == 0)
		return;

	if (elapsedFrames > 1 && LOG_FRAME_DROPS)
	{
		unsigned long oldSyncFrameIndexSinceMidnight = FFrameTimer::GetMidnightRelativeFrameIndex(oldSyncTime, camData->renderFramerate);
		unsigned long repeatFramesCount = elapsedFrames - 1;
		if (repeatFramesCount > 1)
		{
			MIXCAST_WARNING("Experience dropped frames %s - %s for camera %s", 
				*FFrameTimer::GetSMPTE(oldSyncFrameIndexSinceMidnight + 1, camData->renderFramerate),
				*FFrameTimer::GetSMPTE(oldSyncFrameIndexSinceMidnight + repeatFramesCount, camData->renderFramerate),
				*FString(camData->identifier.c_str()));
		}
		else 
		{
			MIXCAST_WARNING("Experience dropped frame %s for camera %s", 
				*FFrameTimer::GetSMPTE(oldSyncFrameIndexSinceMidnight + 1, camData->renderFramerate),
				*FString(camData->identifier.c_str()));
		}
	}

	CaptureFrame(camData);
}
void AExpCamera::CaptureFrame(VirtualCamera* camInfo)
{
	if (camInfo->renderResolutionWidth != LayerCanvas->GetLayerWidth() || camInfo->renderResolutionHeight != LayerCanvas->GetLayerHeight())
	{
		DestroyLayerCanvas();
		CreateLayerCanvas(camInfo);
	}

	UpdateTransform(camInfo);
	UpdateCameraParameters(camInfo);
	ResetHiddenActors();

	FrameStarted.Broadcast(this, camInfo);
	TriggerSceneCaptures(camInfo);
	TriggerFrameComposite();
	FrameEnded.Broadcast(this);
}
void AExpCamera::UpdateTransform(VirtualCamera* camInfo)
{
	FVector camPos = ThriftUtil::ThriftPositionToUnreal(camInfo->currentPosition);
	FQuat camRot = ThriftUtil::ThriftQuatToUnreal(camInfo->currentRotation);
	RootComponent->SetRelativeLocationAndRotation(camPos, camRot);
	RootComponent->UpdateComponentToWorld();
}
void AExpCamera::UpdateCameraParameters(VirtualCamera* camInfo)
{
	//update field of view
	float hFoV = MathUtil::VerticalFoVToHorizontal(camInfo->fieldOfView, (float)camInfo->renderResolutionWidth / camInfo->renderResolutionHeight);
	if (hFoV != FullCapture->FOVAngle) {
		FullCapture->FOVAngle = hFoV;
		if (FullAlphaCapture)
			FullAlphaCapture->FOVAngle = hFoV;

		ForegroundCapture->FOVAngle = hFoV;
		if (ForegroundAlphaCapture)
			ForegroundAlphaCapture->FOVAngle = hFoV;
	}

	//Update foreground clipping plane
	if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::Planar) 
	{
		float playerDist = CalculateForegroundDistance(camInfo);	//meters
		FTransform camTrans = RootComponent->GetComponentTransform();
		FVector camForward = camTrans.TransformVector(FVector::ForwardVector);
		FVector clipPos = camTrans.TransformPosition(FVector::ForwardVector * playerDist);

		ForegroundCapture->ClipPlaneBase = clipPos;
		ForegroundCapture->ClipPlaneNormal = -camForward;
		if (ForegroundAlphaCapture) 
		{
			ForegroundAlphaCapture->ClipPlaneBase = clipPos;
			ForegroundAlphaCapture->ClipPlaneNormal = -camForward;
		}
	}
	else
	{
		ForegroundCutoffMesh->Update(camInfo);
	}

	//Update post-processing settings
	APlayerController* vrPlayer = GEngine->GetFirstLocalPlayerController(GetWorld());
	AActor* camActor = vrPlayer && vrPlayer->PlayerCameraManager ?
		vrPlayer->PlayerCameraManager->ViewTarget.Target :
		nullptr;
	UCameraComponent* camComp = camActor ?
		Cast<UCameraComponent>(camActor->GetComponentByClass(UCameraComponent::StaticClass())) :
		nullptr;

	if (camComp)
	{
		FullCapture->PostProcessSettings = camComp->PostProcessSettings;
		FullCapture->PostProcessBlendWeight = camComp->PostProcessBlendWeight;
		if (ForegroundAlphaCapture) {
			ForegroundCapture->PostProcessSettings = camComp->PostProcessSettings;
			ForegroundCapture->PostProcessBlendWeight = camComp->PostProcessBlendWeight;
		}
	}
}
void AExpCamera::ResetHiddenActors()
{
	FullCapture->HiddenActors.Empty();
	if (FullAlphaCapture)
		FullAlphaCapture->HiddenActors.Empty();
	ForegroundCapture->HiddenActors.Empty();
	if (ForegroundAlphaCapture)
		ForegroundAlphaCapture->HiddenActors.Empty();

	for (auto& cutoffMesh : UForegroundCutoffMeshComponent::Active)
	{
		AActor* owner = cutoffMesh->GetOwner();
		FullCapture->HiddenActors.Add(owner);
		if (FullAlphaCapture)
			FullAlphaCapture->HiddenActors.Add(owner);
		if (cutoffMesh != ForegroundCutoffMesh)
		{
			ForegroundCapture->HiddenActors.Add(owner);
			if (ForegroundAlphaCapture)
				ForegroundAlphaCapture->HiddenActors.Add(owner);
		}
	}
}
void AExpCamera::TriggerSceneCaptures(VirtualCamera* camInfo)
{
	if (camInfo->usesFullRender) {
		if (fullRenderSetAsTransparent != camInfo->isBackgroundTranslucent) {
			SetCaptureTransparency(FullCapture, camInfo->isBackgroundTranslucent);
			fullRenderSetAsTransparent = camInfo->isBackgroundTranslucent;
		}
		FullCapture->CaptureScene();
		if(camInfo->isBackgroundTranslucent && SdkData::GetProjectSettings()->TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
			FullAlphaCapture->CaptureScene();
	}

	if (camInfo->usesForeground)
	{
		ForegroundCapture->CaptureScene();
		if (SdkData::GetProjectSettings()->TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
			ForegroundAlphaCapture->CaptureScene();
	}
}
void AExpCamera::TriggerFrameComposite()
{
	LayerCanvas->FastUpdateResource();
}

float AExpCamera::CalculateForegroundDistance(VirtualCamera* camInfo)
{
	int hmdCount = 0;
	FVector devicePosSum = FVector::ZeroVector;
	for (int i = 0; i < SdkData::TrackedObjects.size(); i++)
	{
		TrackedObject* trackedObject = &SdkData::TrackedObjects[i];

		if (!trackedObject->connected || trackedObject->hideFromUser)
			continue;

		if (trackedObject->objectType != ObjectType::HMD)
			continue;

		FVector objectPosition = ThriftUtil::ThriftPositionToUnreal(trackedObject->position);

		//if (trackedObject->source == TrackingSource::OCULUS)
		//	objectPosition += trackedObject->rotation.unity * VRInfo.OCULUS_RIFT_HMD_TO_HEAD;
		//else if (trackedObject->source == TrackingSource::OPENVR)
		//	objectPosition += trackedObject->rotation.unity * VRInfo.VIVE_HMD_TO_HEAD;

		devicePosSum += objectPosition;
		hmdCount++;
	}

	if (hmdCount == 0)
		return 0;

	devicePosSum /= hmdCount;

	FVector camPos = ThriftUtil::ThriftPositionToUnreal(camInfo->currentPosition);
	FQuat camRot = ThriftUtil::ThriftQuatToUnreal(camInfo->currentRotation);
	FTransform camTrans = FTransform(camRot, camPos);

	FVector playerPosInCameraSpace = camTrans.Inverse().TransformPosition(devicePosSum);
	return playerPosInCameraSpace.X;
}

void AExpCamera::SetCaptureTransparency(USceneCaptureComponent2D* capture, bool isTransparent)
{
	if (isTransparent) {
		capture->ShowFlags.SetFog(false);
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 23
		capture->ShowFlags.SetAtmosphericFog(false);
#endif
	}
}

void AExpCamera::SetActorHidden(AActor* actor)
{
	FullCapture->HiddenActors.Add(actor);
	if (FullAlphaCapture)
		FullAlphaCapture->HiddenActors.Add(actor);
	ForegroundCapture->HiddenActors.Add(actor);
	if (ForegroundAlphaCapture)
		ForegroundAlphaCapture->HiddenActors.Add(actor);
}

float AExpCamera::GetNearClipPlane()
{
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 23
	if (FullCapture->bOverride_CustomNearClippingPlane)
		return FullCapture->CustomNearClippingPlane;
	else
#endif
		return GNearClippingPlane;
}
