/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "SharedTextureReceiver.h"
#include "MixCast.h"
#include "Transmission/Thrift/ServiceCommunicator.h"
#include "Utility/MiscUtil.hpp"

#include "Runtime/Launch/Resources/Version.h"
#include "Engine/Texture2D.h"

using namespace mixcast::data;
using namespace mixcast::thrift;

static ID3D11Device* g_D3D11Device;
ID3D11DeviceContext* g_pImmediateContext = NULL;

SharedTextureReceiver::SharedTextureReceiver(FString texId) :
	Texture(NULL),
	SharedResource(NULL)
{
	TextureId = texId;

	FMixCast* mod = ((FMixCast*)&IMixCast::Get());

	EventHandle = mod->OnExternalTextureUpdated().AddRaw(this, &SharedTextureReceiver::RefreshTextureFromEvent);

	TextureInfo = mod->GetClientConnection()->GetSharedTextureInfo(TextureId);
	if (TextureInfo.handle == 0)
		return;

	RefreshTextureObject();
}

SharedTextureReceiver::~SharedTextureReceiver()
{
	FMixCast* mod = ((FMixCast*)&IMixCast::Get());
	if (mod)
		mod->OnExternalTextureUpdated().Remove(EventHandle);

	DestroyTexture();
	TextureReferenceChanged.Broadcast();
}

void SharedTextureReceiver::EnqueueUpdate()
{
	if (Texture == NULL)
		return;

	ENQUEUE_RENDER_COMMAND(void)(
		[&](FRHICommandListImmediate& RHICmdList)
		{
			ID3D11Texture2D* src = (ID3D11Texture2D*)SharedResource;
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 26
			FTextureResource* dstRes = (FTextureResource*)Texture->Resource;
#else
			FTexture2DResource* dstRes = (FTexture2DResource*)Texture->Resource;
#endif
			if (dstRes == NULL)
				return;
			FTexture2DRHIRef dstRhiRef = dstRes->GetTexture2DRHI();
			if (!dstRhiRef.IsValid())
				return;
			ID3D11Texture2D* dst = (ID3D11Texture2D*)dstRhiRef->GetNativeResource();

			g_pImmediateContext->CopyResource(dst, src);
		});
}

void SharedTextureReceiver::RefreshTextureObject()
{
	DestroyTexture();

	if (TextureInfo.handle == 0 || TextureInfo.width == 0 || TextureInfo.height == 0) {
		TextureReferenceChanged.Broadcast();
		return;
	}

	if (g_D3D11Device == nullptr || g_pImmediateContext == NULL) {
		g_D3D11Device = (ID3D11Device*)GDynamicRHI->RHIGetNativeDevice();
		g_D3D11Device->GetImmediateContext(&g_pImmediateContext);
	}

	HANDLE texHandle = (void*)TextureInfo.handle;

	//MIXCAST_LOG("___Open Shared Resource___");
	HRESULT openResult = g_D3D11Device->OpenSharedResource(texHandle, __uuidof(ID3D11Resource), (void**)&SharedResource);
	if (FAILED(openResult)) {
		MIXCAST_ERROR("--FAIL--___Open Shared Resource [%s]___---", *TextureId);
		return;
	}

	EPixelFormat pixelFormat = PF_A32B32G32R32F;
	switch (TextureInfo.format)
	{
	case 0:
	case 2:
		pixelFormat = PF_A32B32G32R32F;
		break;
	case 10:
		pixelFormat = PF_FloatRGBA;
		break;
	case 41:
		pixelFormat = PF_R32_FLOAT;
		break;
	}

	Texture = UTexture2D::CreateTransient(TextureInfo.width, TextureInfo.height, pixelFormat);
	Texture->AddToRoot();
	Texture->UpdateResource();

	TextureReferenceChanged.Broadcast();
}

void SharedTextureReceiver::RefreshTextureFromEvent(std::string TexId, SharedTex TexInfo)
{
	if (MiscUtil::StdAndFStrAreEqual(TexId, TextureId))
	{
		TextureInfo = TexInfo;
		RefreshTextureObject();
	}
}

void SharedTextureReceiver::DestroyTexture()
{
	//FlushRenderingCommands();
	if (SharedResource)
	{
		SharedResource->Release();
		SharedResource = NULL;
	}

	if (Texture) {
		Texture->RemoveFromRoot();

		if (Texture->Resource)
		{
			BeginReleaseResource(Texture->Resource);
			FlushRenderingCommands();
		}

		Texture->MarkPendingKill();
		Texture = nullptr;
	}
}
